@@ -1,9 +1,8 @@
-// Copyright 2021, 2022 Martin Pool
+// Copyright 2021-2024 Martin Pool
 
 //! Print messages and progress bars on the terminal.
 
 use std::borrow::Cow;
-use std::fmt::Write;
 use std::fs::File;
 use std::io;
 use std::sync::{Arc, Mutex};
@@ -11,16 +10,17 @@ use std::time::{Duration, Instant};
 
 use camino::{Utf8Path, Utf8PathBuf};
 use console::{style, StyledObject};
-
+use humantime::format_duration;
+use nutmeg::Destination;
 use tracing::Level;
 use tracing_subscriber::fmt::MakeWriter;
 use tracing_subscriber::prelude::*;
 
-use crate::outcome::{LabOutcome, SummaryOutcome};
+use crate::options::Colors;
+use crate::outcome::{LabOutcome, ScenarioOutcome, SummaryOutcome};
 use crate::scenario::Scenario;
-use crate::{last_line, Mutant, Options, Phase, Result, ScenarioOutcome};
-
-static COPY_MESSAGE: &str = "Copy source to scratch directory";
+use crate::tail_file::TailFile;
+use crate::{Mutant, Options, Phase};
 
 /// An interface to the console for the rest of cargo-mutants.
 ///
@@ -60,9 +60,9 @@ impl Console {
     }
 
     /// Update that a cargo task is starting.
-    pub fn scenario_started(&self, scenario: &Scenario, log_file: &Utf8Path) {
+    pub fn scenario_started(&self, dir: &Utf8Path, scenario: &Scenario, log_file: File) {
         let start = Instant::now();
-        let scenario_model = ScenarioModel::new(scenario, start, log_file.to_owned());
+        let scenario_model = ScenarioModel::new(dir, scenario, start, log_file);
         self.view.update(|model| {
             model.scenario_models.push(scenario_model);
         });
@@ -71,12 +71,15 @@ impl Console {
     /// Update that cargo finished.
     pub fn scenario_finished(
         &self,
+        dir: &Utf8Path,
         scenario: &Scenario,
         outcome: &ScenarioOutcome,
         options: &Options,
     ) {
         self.view.update(|model| {
-            model.mutants_done += scenario.is_mutant() as usize;
+            if scenario.is_mutant() {
+                model.mutants_done += 1;
+            }
             match outcome.summary() {
                 SummaryOutcome::CaughtMutant => model.mutants_caught += 1,
                 SummaryOutcome::MissedMutant => model.mutants_missed += 1,
@@ -85,7 +88,7 @@ impl Console {
                 SummaryOutcome::Success => model.successes += 1,
                 SummaryOutcome::Failure => model.failures += 1,
             }
-            model.remove_scenario(scenario);
+            model.remove_scenario(dir);
         });
 
         if (outcome.mutant_caught() && !options.print_caught)
@@ -96,14 +99,11 @@ impl Console {
             return;
         }
 
-        let mut s = String::with_capacity(100);
-        write!(
-            s,
-            "{} ... {}",
-            style_scenario(scenario),
-            style_outcome(outcome)
-        )
-        .unwrap();
+        let mut s = format!(
+            "{:8} {}",
+            style_outcome(outcome),
+            style_scenario(scenario, true),
+        );
         if options.show_times {
             let prs: Vec<String> = outcome
                 .phase_results()
@@ -116,7 +116,8 @@ impl Console {
                     )
                 })
                 .collect();
-            let _ = write!(s, " in {}", prs.join(" + "));
+            s.push_str(" in ");
+            s.push_str(&prs.join(" + "));
         }
         if outcome.should_show_logs() || options.show_all_logs {
             s.push('\n');
@@ -128,43 +129,34 @@ impl Console {
             );
         }
         s.push('\n');
-        self.view.message(&s);
+        self.message(&s);
     }
 
-    /// Update that a test timeout was auto-set.
-    pub fn autoset_timeout(&self, timeout: Duration) {
-        self.message(&format!(
-            "Auto-set test timeout to {}\n",
-            style_secs(timeout)
-        ));
-    }
-
-    pub fn build_dirs_start(&self, _n: usize) {
-        // self.message(&format!("Make {n} more build directories...\n"));
-    }
-
-    pub fn build_dirs_finished(&self) {}
-
-    pub fn start_copy(&self) {
+    pub fn start_copy(&self, dir: &Utf8Path) {
         self.view.update(|model| {
-            assert!(model.copy_model.is_none());
-            model.copy_model = Some(CopyModel::new());
+            model.copy_models.push(CopyModel::new(dir.to_owned()));
         });
     }
 
-    pub fn finish_copy(&self) {
+    pub fn finish_copy(&self, dir: &Utf8Path) {
         self.view.update(|model| {
-            model.copy_model = None;
+            let idx = model
+                .copy_models
+                .iter()
+                .position(|m| m.dest == dir)
+                .expect("copy model not found");
+            model.copy_models.swap_remove(idx);
         });
     }
 
-    pub fn copy_progress(&self, total_bytes: u64) {
+    pub fn copy_progress(&self, dest: &Utf8Path, total_bytes: u64) {
         self.view.update(|model| {
             model
-                .copy_model
-                .as_mut()
+                .copy_models
+                .iter_mut()
+                .find(|m| m.dest == dest)
                 .expect("copy in progress")
-                .bytes_copied(total_bytes)
+                .bytes_copied(total_bytes);
         });
     }
 
@@ -178,7 +170,7 @@ impl Console {
         self.view.update(|model| {
             model.n_mutants = n_mutants;
             model.lab_start_time = Some(Instant::now());
-        })
+        });
     }
 
     /// Update that work is starting on testing a given number of mutants.
@@ -188,16 +180,16 @@ impl Console {
     }
 
     /// A new phase of this scenario started.
-    pub fn scenario_phase_started(&self, scenario: &Scenario, phase: Phase) {
+    pub fn scenario_phase_started(&self, dir: &Utf8Path, phase: Phase) {
         self.view.update(|model| {
-            model.find_scenario_mut(scenario).phase_started(phase);
-        })
+            model.find_scenario_mut(dir).phase_started(phase);
+        });
     }
 
-    pub fn scenario_phase_finished(&self, scenario: &Scenario, phase: Phase) {
+    pub fn scenario_phase_finished(&self, dir: &Utf8Path, phase: Phase) {
         self.view.update(|model| {
-            model.find_scenario_mut(scenario).phase_finished(phase);
-        })
+            model.find_scenario_mut(dir).phase_finished(phase);
+        });
     }
 
     pub fn lab_finished(&self, lab_outcome: &LabOutcome, start_time: Instant, options: &Options) {
@@ -211,15 +203,19 @@ impl Console {
     }
 
     pub fn clear(&self) {
-        self.view.clear()
+        self.view.clear();
     }
 
     pub fn message(&self, message: &str) {
-        self.view.message(message)
+        // A workaround for nutmeg not being able to coordinate writes to both stdout and
+        // stderr...
+        // <https://github.com/sourcefrog/nutmeg/issues/11>
+        self.view.clear();
+        print!("{message}");
     }
 
     pub fn tick(&self) {
-        self.view.update(|_| ())
+        self.view.update(|_| ());
     }
 
     /// Return a tracing `MakeWriter` that will send messages via nutmeg to the console.
@@ -242,10 +238,13 @@ impl Console {
 
     /// Configure tracing to send messages to the console and debug log.
     ///
-    /// The debug log is opened later and provided by [Console::set_debug_log].
-    pub fn setup_global_trace(&self, console_trace_level: Level) -> Result<()> {
+    /// The debug log is opened later and provided by [`Console::set_debug_log`].
+    pub fn setup_global_trace(&self, console_trace_level: Level, colors: Colors) {
         // Show time relative to the start of the program.
         let uptime = tracing_subscriber::fmt::time::uptime();
+        let stderr_colors = colors
+            .forced_value()
+            .unwrap_or_else(::console::colors_enabled_stderr);
         let debug_log_layer = tracing_subscriber::fmt::layer()
             .with_ansi(false)
             .with_file(true) // source file name
@@ -254,16 +253,29 @@ impl Console {
             .with_writer(self.make_debug_log_writer());
         let level_filter = tracing_subscriber::filter::LevelFilter::from_level(console_trace_level);
         let console_layer = tracing_subscriber::fmt::layer()
-            .with_ansi(true)
+            .with_ansi(stderr_colors)
             .with_writer(self.make_terminal_writer())
             .with_target(false)
-            .with_timer(uptime)
+            .without_time()
             .with_filter(level_filter);
         tracing_subscriber::registry()
             .with(debug_log_layer)
             .with(console_layer)
             .init();
-        Ok(())
+    }
+}
+
+pub fn enable_console_colors(colors: Colors) {
+    if let Some(colors) = colors.forced_value() {
+        ::console::set_colors_enabled(colors);
+        ::console::set_colors_enabled_stderr(colors);
+    }
+    // Otherwise, let the console crate decide, based on isatty, etc.
+}
+
+impl Default for Console {
+    fn default() -> Self {
+        Self::new()
     }
 }
 
@@ -272,7 +284,7 @@ pub struct TerminalWriter {
     view: Arc<nutmeg::View<LabModel>>,
 }
 
-impl<'w> MakeWriter<'w> for TerminalWriter {
+impl MakeWriter<'_> for TerminalWriter {
     type Writer = Self;
 
     fn make_writer(&self) -> Self::Writer {
@@ -299,7 +311,7 @@ impl std::io::Write for TerminalWriter {
 /// Write trace output to the debug log file if it's open.
 pub struct DebugLogWriter(Arc<Mutex<Option<File>>>);
 
-impl<'w> MakeWriter<'w> for DebugLogWriter {
+impl MakeWriter<'_> for DebugLogWriter {
     type Writer = Self;
 
     fn make_writer(&self) -> Self::Writer {
@@ -332,10 +344,11 @@ impl io::Write for DebugLogWriter {
 #[derive(Default)]
 struct LabModel {
     walk_tree: Option<WalkModel>,
-    copy_model: Option<CopyModel>,
+    /// Copy jobs in progress
+    copy_models: Vec<CopyModel>,
     scenario_models: Vec<ScenarioModel>,
     lab_start_time: Option<Instant>,
-    // The instant when we started trying mutation scenarios, after running the baseline.
+    /// The instant when we started trying mutation scenarios, after running the baseline.
     mutants_start_time: Option<Instant>,
     mutants_done: usize,
     n_mutants: usize,
@@ -348,55 +361,53 @@ struct LabModel {
 }
 
 impl nutmeg::Model for LabModel {
+    #[allow(clippy::cast_precision_loss)]
     fn render(&mut self, width: usize) -> String {
         let mut s = String::with_capacity(1024);
         if let Some(walk_tree) = &mut self.walk_tree {
             s += &walk_tree.render(width);
         }
-        if let Some(copy) = self.copy_model.as_mut() {
-            s.push_str(&copy.render(width));
+        for copy_model in &mut self.copy_models {
+            if !s.is_empty() {
+                s.push('\n');
+            }
+            s.push_str(&copy_model.render(width));
         }
-        if !s.is_empty() {
-            s.push('\n')
+        for sm in &mut self.scenario_models {
+            if !s.is_empty() {
+                s.push('\n');
+            }
+            s.push_str(&sm.render(width));
         }
         if let Some(lab_start_time) = self.lab_start_time {
+            if !s.is_empty() {
+                s.push('\n');
+            }
             let elapsed = lab_start_time.elapsed();
-            let percent = if self.n_mutants > 0 {
-                ((self.mutants_done as f64) / (self.n_mutants as f64) * 100.0).round()
-            } else {
-                0.0
-            };
-            write!(
-                s,
-                "{}/{} mutants tested, {}% done",
+            s += &format!(
+                "{}/{} mutants tested",
                 style(self.mutants_done).cyan(),
                 style(self.n_mutants).cyan(),
-                style(percent).cyan(),
-            )
-            .unwrap();
+            );
             if self.mutants_missed > 0 {
-                write!(
-                    s,
+                s += &format!(
                     ", {} {}",
                     style(self.mutants_missed).cyan(),
-                    style("missed").red()
-                )
-                .unwrap();
+                    style("MISSED").red()
+                );
             }
             if self.timeouts > 0 {
-                write!(
-                    s,
+                s += &format!(
                     ", {} {}",
                     style(self.timeouts).cyan(),
                     style("timeout").red()
-                )
-                .unwrap();
+                );
             }
             if self.mutants_caught > 0 {
-                write!(s, ", {} caught", style(self.mutants_caught).cyan()).unwrap();
+                s.push_str(&format!(", {} caught", style(self.mutants_caught).cyan()));
             }
             if self.unviable > 0 {
-                write!(s, ", {} unviable", style(self.unviable).cyan()).unwrap();
+                s.push_str(&format!(", {} unviable", style(self.unviable).cyan()));
             }
             // Maybe don't report these, because they're uninteresting?
             // if self.successes > 0 {
@@ -405,43 +416,36 @@ impl nutmeg::Model for LabModel {
             // if self.failures > 0 {
             //     write!(s, ", {} failures", self.failures).unwrap();
             // }
-            write!(s, ", {} elapsed", style_minutes_seconds(elapsed)).unwrap();
+            s.push_str(&format!(", {} elapsed", style_duration(elapsed)));
             if self.mutants_done > 2 {
-                write!(
-                    s,
+                let done = self.mutants_done as f64;
+                let remain = self.n_mutants as f64 - done;
+                let mut remaining_secs =
+                    self.mutants_start_time.unwrap().elapsed().as_secs_f64() * remain / done;
+                if remaining_secs > 300.0 {
+                    // Round up to minutes
+                    remaining_secs = ((remaining_secs + 30.0) / 60.0).ceil() * 60.0;
+                }
+                s += &format!(
                     ", about {} remaining",
-                    style(nutmeg::estimate_remaining(
-                        &self.mutants_start_time.unwrap(),
-                        self.mutants_done,
-                        self.n_mutants
-                    ))
-                    .cyan()
-                )
-                .unwrap();
+                    style_duration(Duration::from_secs_f64(remaining_secs.ceil()))
+                );
             }
-            writeln!(s).unwrap();
-        }
-        for sm in self.scenario_models.iter_mut() {
-            s.push_str(&sm.render(width));
-            s.push('\n');
-        }
-        while s.ends_with('\n') {
-            s.pop();
         }
         s
     }
 }
 
 impl LabModel {
-    fn find_scenario_mut(&mut self, scenario: &Scenario) -> &mut ScenarioModel {
+    fn find_scenario_mut(&mut self, dir: &Utf8Path) -> &mut ScenarioModel {
         self.scenario_models
             .iter_mut()
-            .find(|sm| sm.scenario == *scenario)
-            .expect("scenario is in progress")
+            .find(|sm| sm.dir == *dir)
+            .expect("scenario directory not found")
     }
 
-    fn remove_scenario(&mut self, scenario: &Scenario) {
-        self.scenario_models.retain(|sm| sm.scenario != *scenario);
+    fn remove_scenario(&mut self, dir: &Utf8Path) {
+        self.scenario_models.retain(|sm| sm.dir != *dir);
     }
 }
 
@@ -469,23 +473,24 @@ impl nutmeg::Model for WalkModel {
 ///
 /// It draws the command and some description of what scenario is being tested.
 struct ScenarioModel {
-    scenario: Scenario,
+    /// The directory where this is being built: unique across all models.
+    dir: Utf8PathBuf,
     name: Cow<'static, str>,
     phase_start: Instant,
     phase: Option<Phase>,
     /// Previously-executed phases and durations.
     previous_phase_durations: Vec<(Phase, Duration)>,
-    log_file: Utf8PathBuf,
+    log_tail: TailFile,
 }
 
 impl ScenarioModel {
-    fn new(scenario: &Scenario, start: Instant, log_file: Utf8PathBuf) -> ScenarioModel {
+    fn new(dir: &Utf8Path, scenario: &Scenario, start: Instant, log_file: File) -> ScenarioModel {
         ScenarioModel {
-            scenario: scenario.clone(),
-            name: style_scenario(scenario),
+            dir: dir.to_owned(),
+            name: style_scenario(scenario, true),
             phase: None,
             phase_start: start,
-            log_file,
+            log_tail: TailFile::new(log_file),
             previous_phase_durations: Vec::new(),
         }
     }
@@ -505,23 +510,29 @@ impl ScenarioModel {
 
 impl nutmeg::Model for ScenarioModel {
     fn render(&mut self, _width: usize) -> String {
-        let mut s = String::with_capacity(100);
-        write!(s, "{} ... ", self.name).unwrap();
-        let mut prs = self
-            .previous_phase_durations
-            .iter()
-            .map(|(phase, duration)| format!("{} {}", style_secs(*duration), style(phase).dim()))
-            .collect::<Vec<_>>();
+        let mut parts = Vec::new();
         if let Some(phase) = self.phase {
-            prs.push(format!(
-                "{} {}",
-                style_secs(self.phase_start.elapsed()),
-                style(phase).dim()
-            ));
+            parts.push(style(format!("{phase:8}")).bold().cyan().to_string());
         }
-        write!(s, "{}", prs.join(" + ")).unwrap();
-        if let Ok(last_line) = last_line(&self.log_file) {
-            write!(s, "\n    {}", style(last_line).dim()).unwrap();
+        parts.push(self.name.to_string());
+        parts.push("...".to_string());
+        parts.push(style_secs(self.phase_start.elapsed()).to_string());
+        // let mut prs = self
+        //     .previous_phase_durations
+        //     .iter()
+        //     .map(|(phase, duration)| format!("{} {}", style_secs(*duration), style(phase).dim()))
+        //     .collect::<Vec<_>>();
+        // if prs.len() > 1 {
+        //     prs.insert(0, String::new())
+        // }
+        // parts.push(prs.join(" + "));
+        let mut s = parts.join(" ");
+        if let Ok(last_line) = self.log_tail.last_line() {
+            s.push_str(&format!(
+                "\n{:8} {}",
+                style("└").cyan(),
+                style(last_line).dim()
+            ));
         }
         s
     }
@@ -529,14 +540,15 @@ impl nutmeg::Model for ScenarioModel {
 
 /// A Nutmeg model for progress in copying a tree.
 struct CopyModel {
+    dest: Utf8PathBuf,
     bytes_copied: u64,
     start: Instant,
 }
 
 impl CopyModel {
-    #[allow(dead_code)]
-    fn new() -> CopyModel {
+    fn new(dest: Utf8PathBuf) -> CopyModel {
         CopyModel {
+            dest,
             start: Instant::now(),
             bytes_copied: 0,
         }
@@ -545,32 +557,33 @@ impl CopyModel {
     /// Update that some bytes have been copied.
     ///
     /// `bytes_copied` is the total bytes copied so far.
-    #[allow(dead_code)]
     fn bytes_copied(&mut self, bytes_copied: u64) {
-        self.bytes_copied = bytes_copied
+        self.bytes_copied = bytes_copied;
     }
 }
 
 impl nutmeg::Model for CopyModel {
     fn render(&mut self, _width: usize) -> String {
         format!(
-            "{} ... {} in {}",
-            COPY_MESSAGE,
+            "{:8} {} in {}",
+            style("copy").cyan(),
             style_mb(self.bytes_copied),
-            style_elapsed_secs(self.start),
+            style_secs(self.start.elapsed()),
         )
     }
 }
 
 fn nutmeg_options() -> nutmeg::Options {
-    nutmeg::Options::default().print_holdoff(Duration::from_millis(50))
+    nutmeg::Options::default()
+        .print_holdoff(Duration::from_millis(50))
+        .destination(Destination::Stderr)
 }
 
 /// Return a styled string reflecting the moral value of this outcome.
 pub fn style_outcome(outcome: &ScenarioOutcome) -> StyledObject<&'static str> {
     match outcome.summary() {
         SummaryOutcome::CaughtMutant => style("caught").green(),
-        SummaryOutcome::MissedMutant => style("NOT CAUGHT").red().bold(),
+        SummaryOutcome::MissedMutant => style("MISSED").red().bold(),
         SummaryOutcome::Failure => style("FAILED").red().bold(),
         SummaryOutcome::Success => style("ok").green(),
         SummaryOutcome::Unviable => style("unviable").blue(),
@@ -578,45 +591,18 @@ pub fn style_outcome(outcome: &ScenarioOutcome) -> StyledObject<&'static str> {
     }
 }
 
-pub(crate) fn style_mutant(mutant: &Mutant) -> String {
-    // This is like `impl Display for Mutant`, but with colors.
-    // The text content should be the same.
-    let mut s = String::with_capacity(200);
-    write!(
-        &mut s,
-        "{}:{}",
-        mutant.source_file.tree_relative_slashes(),
-        mutant.span.start.line,
-    )
-    .unwrap();
-    s.push_str(": replace ");
-    s.push_str(&style(mutant.function_name()).bright().magenta().to_string());
-    if !mutant.return_type().is_empty() {
-        s.push(' ');
-        s.push_str(&style(mutant.return_type()).magenta().to_string());
-    }
-    s.push_str(" with ");
-    s.push_str(&style(mutant.replacement_text()).yellow().to_string());
-    s
-}
-
-fn style_elapsed_secs(since: Instant) -> String {
-    style_secs(since.elapsed())
-}
-
 fn style_secs(duration: Duration) -> String {
     style(format!("{:.1}s", duration.as_secs_f32()))
         .cyan()
         .to_string()
 }
 
-fn style_minutes_seconds(duration: Duration) -> String {
-    style(duration_minutes_seconds(duration)).cyan().to_string()
-}
-
-pub fn duration_minutes_seconds(duration: Duration) -> String {
-    let secs = duration.as_secs();
-    format!("{}:{:02}", secs / 60, secs % 60)
+fn style_duration(duration: Duration) -> String {
+    // We don't want silly precision.
+    let duration = Duration::from_secs(duration.as_secs());
+    style(format_duration(duration).to_string())
+        .cyan()
+        .to_string()
 }
 
 fn format_mb(bytes: u64) -> String {
@@ -627,10 +613,10 @@ fn style_mb(bytes: u64) -> StyledObject<String> {
     style(format_mb(bytes)).cyan()
 }
 
-pub fn style_scenario(scenario: &Scenario) -> Cow<'static, str> {
+pub fn style_scenario(scenario: &Scenario, line_col: bool) -> Cow<'static, str> {
     match scenario {
-        Scenario::Baseline => "Unmutated baseline".into(),
-        Scenario::Mutant(mutant) => style_mutant(mutant).into(),
+        Scenario::Baseline => Cow::Borrowed("Unmutated baseline"),
+        Scenario::Mutant(mutant) => Cow::Owned(mutant.to_styled_string(line_col)),
     }
 }
 
@@ -641,20 +627,3 @@ pub fn plural(n: usize, noun: &str) -> String {
         format!("{n} {noun}s")
     }
 }
-
-#[cfg(test)]
-mod test {
-    use super::*;
-    use std::time::Duration;
-
-    #[test]
-    fn test_duration_minutes_seconds() {
-        assert_eq!(duration_minutes_seconds(Duration::ZERO), "0:00");
-        assert_eq!(duration_minutes_seconds(Duration::from_secs(3)), "0:03");
-        assert_eq!(duration_minutes_seconds(Duration::from_secs(73)), "1:13");
-        assert_eq!(
-            duration_minutes_seconds(Duration::from_secs(6003)),
-            "100:03"
-        );
-    }
-}
