@@ -1,7 +1,9 @@
-// Copyright 2021-2023 Martin Pool
+// Copyright 2021-2024 Martin Pool
 
 //! Mutations of replacing a function body with a value of a (hopefully) appropriate type.
 
+#![warn(clippy::pedantic)]
+
 use std::iter;
 
 use itertools::Itertools;
@@ -13,21 +15,21 @@ use syn::{
 };
 use tracing::trace;
 
+use crate::pretty::ToPrettyString;
+
 /// Generate replacement text for a function based on its return type.
 pub(crate) fn return_type_replacements(
     return_type: &ReturnType,
     error_exprs: &[Expr],
-) -> impl Iterator<Item = TokenStream> {
+) -> Vec<TokenStream> {
     match return_type {
         ReturnType::Default => vec![quote! { () }],
         ReturnType::Type(_rarrow, type_) => type_replacements(type_, error_exprs).collect_vec(),
     }
-    .into_iter()
 }
 
 /// Generate some values that we hope are reasonable replacements for a type.
-///
-/// This is really the heart of cargo-mutants.
+#[allow(clippy::too_many_lines)]
 fn type_replacements(type_: &Type, error_exprs: &[Expr]) -> impl Iterator<Item = TokenStream> {
     // This could probably change to run from some configuration rather than
     // hardcoding various types, which would make it easier to support tree-specific
@@ -115,6 +117,17 @@ fn type_replacements(type_: &Type, error_exprs: &[Expr]) -> impl Iterator<Item =
                         quote! { #collection_type::from_iter([#rep]) }
                     }))
                     .collect_vec()
+            } else if let Some((collection_type, key_type, value_type)) = known_map(path) {
+                let key_reps = type_replacements(key_type, error_exprs).collect_vec();
+                let val_reps = type_replacements(value_type, error_exprs).collect_vec();
+                iter::once(quote! { #collection_type::new() })
+                    .chain(
+                        key_reps
+                            .iter()
+                            .cartesian_product(val_reps)
+                            .map(|(k, v)| quote! { #collection_type::from_iter([(#k, #v)]) }),
+                    )
+                    .collect_vec()
             } else if let Some((collection_type, inner_type)) = maybe_collection_or_container(path)
             {
                 // Something like `T<A>` or `T<'a, A>`, when we don't know exactly how
@@ -130,7 +143,10 @@ fn type_replacements(type_: &Type, error_exprs: &[Expr]) -> impl Iterator<Item =
                     }))
                     .collect_vec()
             } else {
-                trace!(?type_, "Return type is not recognized, trying Default");
+                trace!(
+                    type_ = type_.to_pretty_string(),
+                    "Return type is not recognized, trying Default"
+                );
                 vec![quote! { Default::default() }]
             }
         }
@@ -226,22 +242,25 @@ fn type_replacements(type_: &Type, error_exprs: &[Expr]) -> impl Iterator<Item =
 }
 
 fn path_ends_with(path: &Path, ident: &str) -> bool {
-    path.segments.last().map_or(false, |s| s.ident == ident)
+    path.segments.last().is_some_and(|s| s.ident == ident)
 }
 
 fn match_impl_iterator(TypeImplTrait { bounds, .. }: &TypeImplTrait) -> Option<&Type> {
     for bound in bounds {
         if let TypeParamBound::Trait(TraitBound { path, .. }) = bound {
-            if path.segments.len() == 1 && path.segments[0].ident == "Iterator" {
-                if let PathArguments::AngleBracketed(AngleBracketedGenericArguments {
-                    args, ..
-                }) = &path.segments[0].arguments
-                {
-                    if let Some(GenericArgument::AssocType(AssocType { ident, ty, .. })) =
-                        args.first()
+            if let Some(last_segment) = path.segments.last() {
+                if last_segment.ident == "Iterator" {
+                    if let PathArguments::AngleBracketed(AngleBracketedGenericArguments {
+                        args,
+                        ..
+                    }) = &last_segment.arguments
                     {
-                        if ident == "Item" {
-                            return Some(ty);
+                        if let Some(GenericArgument::AssocType(AssocType { ident, ty, .. })) =
+                            args.first()
+                        {
+                            if ident == "Item" {
+                                return Some(ty);
+                            }
                         }
                     }
                 }
@@ -278,6 +297,8 @@ fn known_container(path: &Path) -> Option<(&Ident, &Type)> {
 
 /// Match known simple collections that can be empty or constructed from an
 /// iterator.
+///
+/// Returns the short name (like `"VecDeque"`) and the inner type.
 fn known_collection(path: &Path) -> Option<(&Ident, &Type)> {
     let last = path.segments.last()?;
     if ![
@@ -306,6 +327,26 @@ fn known_collection(path: &Path) -> Option<(&Ident, &Type)> {
     None
 }
 
+/// Match known key-value maps that can be empty or constructed from pair of
+/// recursively-generated values.
+fn known_map(path: &Path) -> Option<(&Ident, &Type, &Type)> {
+    let last = path.segments.last()?;
+    if !["BTreeMap", "HashMap"].iter().any(|v| last.ident == v) {
+        return None;
+    }
+    if let PathArguments::AngleBracketed(AngleBracketedGenericArguments { args, .. }) =
+        &last.arguments
+    {
+        // TODO: Skip lifetime args.
+        // TODO: Return the path with args stripped out.
+        if let Some((GenericArgument::Type(ref key_type), GenericArgument::Type(ref value_type))) =
+            args.iter().collect_tuple()
+        {
+            return Some((&last.ident, key_type, value_type));
+        }
+    }
+    None
+}
 /// Match a type with one type argument, which might be a container or collection.
 fn maybe_collection_or_container(path: &Path) -> Option<(&Ident, &Type)> {
     let last = path.segments.last()?;
@@ -378,7 +419,7 @@ fn path_is_nonzero_unsigned(path: &Path) -> bool {
 /// If this is a path ending in `expected_ident`, return the first type argument, ignoring
 /// lifetimes.
 fn match_first_type_arg<'p>(path: &'p Path, expected_ident: &str) -> Option<&'p Type> {
-    // TODO: Maybe match only things witn one arg?
+    // TODO: Maybe match only things with one arg?
     let last = path.segments.last()?;
     if last.ident == expected_ident {
         if let PathArguments::AngleBracketed(AngleBracketedGenericArguments { args, .. }) =
@@ -402,14 +443,15 @@ mod test {
     use pretty_assertions::assert_eq;
     use syn::{parse_quote, Expr, ReturnType};
 
+    use crate::fnvalue::match_impl_iterator;
     use crate::pretty::ToPrettyString;
 
-    use super::return_type_replacements;
+    use super::{known_map, return_type_replacements};
 
     #[test]
     fn recurse_into_result_bool() {
         check_replacements(
-            parse_quote! {-> std::result::Result<bool> },
+            &parse_quote! {-> std::result::Result<bool> },
             &[],
             &["Ok(true)", "Ok(false)"],
         );
@@ -418,7 +460,7 @@ mod test {
     #[test]
     fn recurse_into_result_result_bool_with_error_values() {
         check_replacements(
-            parse_quote! {-> std::result::Result<Result<bool>> },
+            &parse_quote! {-> std::result::Result<Result<bool>> },
             &[parse_quote! { anyhow!("mutated") }],
             &[
                 "Ok(Ok(true))",
@@ -431,43 +473,43 @@ mod test {
 
     #[test]
     fn u16_replacements() {
-        check_replacements(parse_quote! { -> u16 }, &[], &["0", "1"]);
+        check_replacements(&parse_quote! { -> u16 }, &[], &["0", "1"]);
     }
 
     #[test]
     fn isize_replacements() {
-        check_replacements(parse_quote! { -> isize }, &[], &["0", "1", "-1"]);
+        check_replacements(&parse_quote! { -> isize }, &[], &["0", "1", "-1"]);
     }
 
     #[test]
     fn nonzero_integer_replacements() {
         check_replacements(
-            parse_quote! { -> std::num::NonZeroIsize },
+            &parse_quote! { -> std::num::NonZeroIsize },
             &[],
             &["1", "-1"],
         );
 
-        check_replacements(parse_quote! { -> std::num::NonZeroUsize }, &[], &["1"]);
+        check_replacements(&parse_quote! { -> std::num::NonZeroUsize }, &[], &["1"]);
 
-        check_replacements(parse_quote! { -> std::num::NonZeroU32 }, &[], &["1"]);
+        check_replacements(&parse_quote! { -> std::num::NonZeroU32 }, &[], &["1"]);
     }
 
     #[test]
     fn unit_replacement() {
-        check_replacements(parse_quote! { -> () }, &[], &["()"]);
+        check_replacements(&parse_quote! { -> () }, &[], &["()"]);
     }
 
     #[test]
     fn result_unit_replacement() {
-        check_replacements(parse_quote! { -> Result<(), Error> }, &[], &["Ok(())"]);
+        check_replacements(&parse_quote! { -> Result<(), Error> }, &[], &["Ok(())"]);
 
-        check_replacements(parse_quote! { -> Result<()> }, &[], &["Ok(())"]);
+        check_replacements(&parse_quote! { -> Result<()> }, &[], &["Ok(())"]);
     }
 
     #[test]
     fn http_response_replacement() {
         check_replacements(
-            parse_quote! { -> HttpResponse },
+            &parse_quote! { -> HttpResponse },
             &[],
             &["HttpResponse::Ok().finish()"],
         );
@@ -476,7 +518,7 @@ mod test {
     #[test]
     fn option_usize_replacement() {
         check_replacements(
-            parse_quote! { -> Option<usize> },
+            &parse_quote! { -> Option<usize> },
             &[],
             &["None", "Some(0)", "Some(1)"],
         );
@@ -485,7 +527,7 @@ mod test {
     #[test]
     fn box_usize_replacement() {
         check_replacements(
-            parse_quote! { -> Box<usize> },
+            &parse_quote! { -> Box<usize> },
             &[],
             &["Box::new(0)", "Box::new(1)"],
         );
@@ -494,7 +536,7 @@ mod test {
     #[test]
     fn box_unrecognized_type_replacement() {
         check_replacements(
-            parse_quote! { -> Box<MyObject> },
+            &parse_quote! { -> Box<MyObject> },
             &[],
             &["Box::new(Default::default())"],
         );
@@ -503,7 +545,7 @@ mod test {
     #[test]
     fn vec_string_replacement() {
         check_replacements(
-            parse_quote! { -> std::vec::Vec<String> },
+            &parse_quote! { -> std::vec::Vec<String> },
             &[],
             &["vec![]", "vec![String::new()]", r#"vec!["xyzzy".into()]"#],
         );
@@ -511,18 +553,18 @@ mod test {
 
     #[test]
     fn float_replacement() {
-        check_replacements(parse_quote! { -> f32 }, &[], &["0.0", "1.0", "-1.0"]);
+        check_replacements(&parse_quote! { -> f32 }, &[], &["0.0", "1.0", "-1.0"]);
     }
 
     #[test]
     fn ref_replacement_recurses() {
-        check_replacements(parse_quote! { -> &bool }, &[], &["&true", "&false"]);
+        check_replacements(&parse_quote! { -> &bool }, &[], &["&true", "&false"]);
     }
 
     #[test]
     fn array_replacement() {
         check_replacements(
-            parse_quote! { -> [u8; 256] },
+            &parse_quote! { -> [u8; 256] },
             &[],
             &["[0; 256]", "[1; 256]"],
         );
@@ -533,7 +575,7 @@ mod test {
         // Also checks that it matches the path, even using an atypical path.
         // TODO: Ideally this would be fully qualified like `alloc::sync::Arc::new(String::new())`.
         check_replacements(
-            parse_quote! { -> alloc::sync::Arc<String> },
+            &parse_quote! { -> alloc::sync::Arc<String> },
             &[],
             &["Arc::new(String::new())", r#"Arc::new("xyzzy".into())"#],
         );
@@ -544,16 +586,79 @@ mod test {
         // Also checks that it matches the path, even using an atypical path.
         // TODO: Ideally this would be fully qualified like `alloc::sync::Rc::new(String::new())`.
         check_replacements(
-            parse_quote! { -> alloc::sync::Rc<String> },
+            &parse_quote! { -> alloc::sync::Rc<String> },
             &[],
             &["Rc::new(String::new())", r#"Rc::new("xyzzy".into())"#],
         );
     }
 
+    #[test]
+    fn match_known_collection() {
+        assert_eq!(
+            super::known_collection(&parse_quote! { std::collections::VecDeque<String> }),
+            Some((&parse_quote! { VecDeque }, &parse_quote! { String }))
+        );
+
+        assert_eq!(
+            super::known_collection(&parse_quote! { std::collections::BinaryHeap<(u32, u32)> }),
+            Some((&parse_quote! { BinaryHeap }, &parse_quote! { (u32, u32) }))
+        );
+
+        assert_eq!(
+            super::known_collection(&parse_quote! { LinkedList<[u8; 256]> }),
+            Some((&parse_quote! { LinkedList }, &parse_quote! { [u8; 256] }))
+        );
+
+        assert_eq!(super::known_collection(&parse_quote! { Arc<String> }), None);
+
+        // This might be a collection, and is handled generically, but it's not a specifically known
+        // collection type. (Maybe we shouldn't bother knowing specific types?)
+        assert_eq!(
+            super::known_collection(&parse_quote! { Wibble<&str> }),
+            None
+        );
+    }
+
+    #[test]
+    fn match_known_map() {
+        assert_eq!(
+            super::known_map(&parse_quote! { std::collections::BTreeMap<String, usize> }),
+            Some((
+                &parse_quote! { BTreeMap },
+                &parse_quote! { String },
+                &parse_quote! { usize }
+            ))
+        );
+
+        assert_eq!(
+            super::known_map(&parse_quote! { std::collections::HashMap<(usize, usize), bool> }),
+            Some((
+                &parse_quote! { HashMap },
+                &parse_quote! { (usize, usize) },
+                &parse_quote! { bool }
+            ))
+        );
+
+        assert_eq!(
+            super::known_map(&parse_quote! { Option<(usize, usize)> }),
+            None
+        );
+
+        assert_eq!(
+            super::known_map(&parse_quote! { MyMap<String, usize> }),
+            None,
+        );
+
+        assert_eq!(
+            super::known_map(&parse_quote! { Pair<String, usize> }),
+            None,
+        );
+    }
+
     #[test]
     fn btreeset_replacement() {
         check_replacements(
-            parse_quote! { -> std::collections::BTreeSet<String> },
+            &parse_quote! { -> std::collections::BTreeSet<String> },
             &[],
             &[
                 "BTreeSet::new()",
@@ -566,7 +671,7 @@ mod test {
     #[test]
     fn cow_generates_borrowed_and_owned() {
         check_replacements(
-            parse_quote! { -> Cow<'static, str> },
+            &parse_quote! { -> Cow<'static, str> },
             &[],
             &[
                 r#"Cow::Borrowed("")"#,
@@ -580,9 +685,9 @@ mod test {
     #[test]
     fn unknown_container_replacement() {
         // This looks like something that holds a &str, and maybe can be constructed
-        // from a &str, but we don't know anythig else about it, so we just guess.
+        // from a &str, but we don't know anything else about it, so we just guess.
         check_replacements(
-            parse_quote! { -> UnknownContainer<'static, str> },
+            &parse_quote! { -> UnknownContainer<'static, str> },
             &[],
             &[
                 "UnknownContainer::new()",
@@ -599,16 +704,16 @@ mod test {
     #[test]
     fn tuple_combinations() {
         check_replacements(
-            parse_quote! { -> (bool, usize) },
+            &parse_quote! { -> (bool, usize) },
             &[],
             &["(true, 0)", "(true, 1)", "(false, 0)", "(false, 1)"],
-        )
+        );
     }
 
     #[test]
     fn tuple_combination_longer() {
         check_replacements(
-            parse_quote! { -> (bool, Option<String>) },
+            &parse_quote! { -> (bool, Option<String>) },
             &[],
             &[
                 "(true, None)",
@@ -618,13 +723,13 @@ mod test {
                 "(false, Some(String::new()))",
                 r#"(false, Some("xyzzy".into()))"#,
             ],
-        )
+        );
     }
 
     #[test]
     fn iter_replacement() {
         check_replacements(
-            parse_quote! { -> impl Iterator<Item = String> },
+            &parse_quote! { -> impl Iterator<Item = String> },
             &[],
             &[
                 "::std::iter::empty()",
@@ -634,10 +739,29 @@ mod test {
         );
     }
 
+    #[test]
+    fn impl_matches_iterator() {
+        assert_eq!(
+            match_impl_iterator(&parse_quote! { impl std::iter::Iterator<Item = String> }),
+            Some(&parse_quote! { String })
+        );
+        assert_eq!(
+            match_impl_iterator(&parse_quote! { impl Iterator<Item = String> }),
+            Some(&parse_quote! { String })
+        );
+        // Strange, maybe it's a type defined in this crate, but we don't know what to
+        // do with it.
+        assert_eq!(match_impl_iterator(&parse_quote! { impl Iterator }), None);
+        assert_eq!(
+            match_impl_iterator(&parse_quote! { impl Borrow<String> }),
+            None
+        );
+    }
+
     #[test]
     fn slice_replacement() {
         check_replacements(
-            parse_quote! { -> [u8] },
+            &parse_quote! { -> [u8] },
             &[],
             &[
                 "Vec::leak(Vec::new())",
@@ -647,12 +771,35 @@ mod test {
         );
     }
 
-    fn check_replacements(return_type: ReturnType, error_exprs: &[Expr], expected: &[&str]) {
+    #[test]
+    fn btreemap_replacement() {
+        check_replacements(
+            &parse_quote! { -> BTreeMap<String, bool> },
+            &[],
+            &[
+                "BTreeMap::new()",
+                "BTreeMap::from_iter([(String::new(), true)])",
+                "BTreeMap::from_iter([(String::new(), false)])",
+                "BTreeMap::from_iter([(\"xyzzy\".into(), true)])",
+                "BTreeMap::from_iter([(\"xyzzy\".into(), false)])",
+            ],
+        );
+    }
+
+    fn check_replacements(return_type: &ReturnType, error_exprs: &[Expr], expected: &[&str]) {
         assert_eq!(
-            return_type_replacements(&return_type, error_exprs)
+            return_type_replacements(return_type, error_exprs)
+                .into_iter()
                 .map(|t| t.to_pretty_string())
                 .collect_vec(),
             expected
         );
     }
+
+    #[test]
+    fn match_map() {
+        assert!(known_map(&parse_quote! { BTreeMap<String, usize> }).is_some());
+        assert!(known_map(&parse_quote! { HashMap<(usize, usize), bool> }).is_some());
+        assert!(known_map(&parse_quote! { Option<(usize, usize)> }).is_none());
+    }
 }
